/* Copyright (C) 2008 GFRN systems

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The latest version of this program may be found at
   http://CQiNet.sourceforge.net

   $Log: voip.cpp,v $
   Revision 1.5  2008/04/24 23:59:13  wb6ymh
   Added callto CheckDtmfTimeouts to PollTx.

   Revision 1.4  2008/04/11 16:48:40  wb6ymh
   Modified GetRxAudio to clear CToneTimer before calling SendCtone.  Fixes
   infinite loop crash under certain rare timing conditions).

   Revision 1.3  2008/03/12 21:44:41  wb6ymh
   Modified VoipNode::VoipNode to make a copy of the Callsign rather than
   saving a pointer.  The address keeps changing as SDES packets are processed
   by conference.c.

   Revision 1.2  2008/03/11 14:21:43  wb6ymh
   1. Moved include for common.h to after system headers.  Fixes compile
      problems (min macro) on some systems.
   2. Added ALOG macro and logging in numerious places to track audio processing.
      Enabled by DLOG_VOIP_AUDIO bit (0x10).
   3. Added BufNum to NetAudio structure to track transmitter's current position
      in the receive audio stream.  Fixes bug that caused one transmitter to
      consume a receiver's audio N times when the receiver was linked to N
      transmitters.
   4. Corrected a bug in GetRxAudio which caused a crash when the CToneTimer was
      started when only a single buffer was queued for consumption.  This
      occurs when there's packet loss.

   Revision 1.1  2008/03/09 17:10:14  wb6ymh
   Inital import - files broken out of old audio.cpp and audio.h due to
   major restructuring of the internal infrastructure to treat VoIP connections
   as pseudo ports.

*/ 
// FreeBSD, Linux, etc..
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#ifdef TIME_WITH_SYS_TIME
   #include <sys/time.h>
   #include <time.h>
#else
   #ifdef HAVE_SYS_TIME_H
      #include <sys/time.h>
   #else
      #include <time.h>
   #endif
#endif
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <string.h>
#include <sys/soundcard.h>
#ifdef HAVE_FCNTL_H
   #include <fcntl.h>
#endif
#include <netdb.h>
#include <ctype.h>
#ifdef HAVE_SYS_TIMEB_H
   #include <sys/timeb.h>
#endif
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <math.h>
#include <typeinfo>

#ifdef HAVE_SYS_SYSLIMITS_H
#include <sys/syslimits.h> // for PATH_MAX
#endif

#include <sys/file.h>

#include "common.h"
#include "rtp.h"
#include "avl.h"
#include "main.h"
#include "configvars.h"
#include "configfile.h"
#include "dirclient.h"
#include "conference.h"
#include "users.h"
#include "eventhook.h"
#include "ctcss.h"
#include "tbd.h"
#include "devinput.h"
#include "linkbox.h"
#include "nodes.h"
#include "ports.h"
#include "voip.h"

#ifdef USE_DMALLOC
#include "dmalloc.h"
#endif

#define ALOG(x) if(Debug & DLOG_VOIP_AUDIO) LogErr x

/*
   return if a <= b < c in a circular sense.
   (From Tanenbaum 1981, pg 158)
*/
int Between(u_int16 a, u_int16 b, u_int16 c)
{
   if( (a <= b && b < c) || (c < a && a <= b) || (b < c && c < a) ) {
      return TRUE;
   }
   return FALSE;
}

// Called by "conference" code when an rtp event occurs
int EndPoint(ClientInfo *p,ConfClient *pCC,EventType Event)
{
   RtpData *pRtpData;
   VoipNode *pVoipNode = (VoipNode *) pCC->pVoipNode;

   if(pVoipNode != NULL || Event == EVENT_INIT) {
      switch(Event) {
         case EVENT_INIT:
            VoipNode::VoipInit(pCC);
            break;

         case EVENT_RTP_RX:
           if(pCC->bInConf && !pCC->bMuted && pCC->bTalking) {
               if(pCC->Proto == PROTO_ILINK || pCC->Proto == PROTO_RTP) {
                  pRtpData = (RtpData *) malloc(sizeof(RtpData));
                  if(pRtpData != NULL) {
                     pRtpData->pData = (rtp_hdr_t *) p->Buf;
                     pRtpData->Len = p->Count;
                     pRtpData->Link = NULL;
                     p->Buf = (char *) malloc(p->BufSize);
                  // For debug convenience convert the sequence number into 
                  // host order
                     pRtpData->pData->seq = ntohs(pRtpData->pData->seq);
                  }
                  pVoipNode->QueueRTPData(pRtpData);
                  DFLUSH();
               }
               else if(pCC->Proto == PROTO_SF) {
                  pVoipNode->QueueSFData((soundbuf *) p->Buf);
               }
               else {
               // This shouldn't happen !
                  CORE_DUMP();
               }
            }
            break;

         case EVENT_RTP_TO:
            D2PRINTF(("e"));
            pVoipNode->StopTalking();
            DFLUSH();
            break;

         case EVENT_RTCP_RX:
         case EVENT_RTCP_TO:
            break;

         case EVENT_DELETE:
         // cleanup, this guy is going to be history shortly.
#if 0
            if(fp != NULL) {
               fclose(fp);
               fp = NULL;
            }
#endif
            EndPointCleanup(pCC);
            break;
      }
   }
   else {
      LOG_ERROR(("%s: pVoipNode == NULL, Event: %d, client: %s.\n",__FUNCTION__,
                 Event,pCC->Callsign));
   }
   return FALSE;
}

VoipNode::VoipNode(ConfClient *pCC)
{
   mCC = pCC;
   NodeName = strdup(pCC->Callsign);
   ADPCM_Bytes = 0;
   AudioQCount = 0;
   bRxActive = FALSE;
   bRxBuffering = FALSE;
   bRxIdle = TRUE;
   DtmfCTCSSMode = 0;
   GsmFramesAvail = 0;
   gsmh = NULL;
   hGSM_Encode = NULL;
   LatePackets = 0;
   MissedPackets = 0;
   NextSeq = 0;
   OutofSeqPackets = 0;
   pInHead = NULL;
   pRtpQueue = NULL;
   RtpQCount = 0;
   RtpTxSequence = 0;
   TxSequence = 0;

   avl_insert(NodeTree,this);
}

void VoipNode::QueueAudio(NetAudio *pAudio) 
{
   if(DtmfMethod == 1) {
      dtmf_rx(&DtmfRxState,(const int16_t *) pAudio->Buf,pAudio->Available/2);
   }

   if(pAudio != NULL) {
      if(CtoneTimer.tv_sec != 0) {
      // Ctone timer is running, advance it
         Add2CtoneTimer(pAudio->Available);
      }
#if 0
      if(fp != NULL) {
         fwrite(pAudio->Buf,pAudio->Available,1,fp);
      }
#endif
      if(Transmitters > 0) {
         AudioQCount++;
         ALOG(("%s Queuing BufNum %d, AudioQCount: %d\n",NodeName,
               pAudio->BufNum,AudioQCount));
         pAudio->Link = NULL;
         pAudio->Users = Transmitters;
         if(pInHead == NULL) {
            pInHead = pAudio;
         }
         else {
            pInTail->Link = pAudio;
         }
         pInTail = pAudio;
      }
      else {
      // If a tree falls in the forest and nobody is listening ...
         ALOG(("%s No transmitters, freeing BufNum %d, AudioQCount: %d\n",
               NodeName,pAudio->BufNum,AudioQCount));
         free(pAudio);
      }
   }
}

NetAudio *VoipNode::AllocAudioBuf(int Samples)
{
   NetAudio *pAudio;
    
   pAudio = (NetAudio *) malloc(sizeof(NetAudio) + Samples * sizeof(int16));
   if(pAudio != NULL) {
      pAudio->BufNum = BufNum++;
      pAudio->Link = NULL;
   }
   return pAudio;
}

void VoipNode::DecompressGSM(gsm_byte *pGSM,int Len,gsm gsmh)
{
   int i;
   int GsmFrames = 0;
   NetAudio *pAudio = NULL;
   int AudioSamples = 0;
   gsm_signal *pDecoded;
   
   GsmFrames = Len / sizeof(gsm_frame);
   AudioSamples = GsmFrames * AUDIO_SAMPLES_PER_GSM_FRAME;

   if(AudioSamples > 0) {
      if((pAudio = AllocAudioBuf(AudioSamples)) == NULL) {
         LOG_ERROR(("%s: AllocAudioBuf failed.\n",__FUNCTION__));
      }
      else {
         pAudio->Available = AudioSamples * sizeof(int16);
         pDecoded = (gsm_signal *) pAudio->Buf;
         for(i = 0; i < GsmFrames; i++) {
            gsm_decode(gsmh,pGSM,pDecoded);
            pGSM += sizeof(gsm_frame);
            pDecoded += AUDIO_SAMPLES_PER_GSM_FRAME;
         }
         QueueAudio(pAudio);
      }
   }
}

void VoipNode::DecompressADPCM(char *p,int Len,struct adpcm_state *pState)
{
   int AudioSamples = Len * 2;
   NetAudio *pAudio = NULL;

   if(AudioSamples > 0) {
      if((pAudio = AllocAudioBuf(AudioSamples)) == NULL) {
         LOG_ERROR(("%s: AllocAudioBuf failed.\n",__FUNCTION__));
      }
      else {
         pAudio->Available = AudioSamples * 2;
         adpcm_decoder(p,(short *)pAudio->Buf,AudioSamples,pState);
         QueueAudio(pAudio);
      }
   }
}

// decompress as many rtp packets as possible
void VoipNode::DecompressRTP()
{
   RtpData *pRtp = pRtpQueue;
   RtpData *pNext;
   gsm_byte *pGSM;
   int DataLen;
   char *dp;
   struct adpcm_state adpcm;      /* ADPCM compression state */
   u_int8 *sp;
   
   while(pRtp != NULL && NextSeq == pRtp->pData->seq) {
      NextSeq++;
      pRtpQueue = pRtp->Link;
      RtpQCount--;
      switch(pRtp->pData->pt) {
         case RTP_PT_GSM:
            if(gsmh == NULL) {
               gsmh = gsm_create();
            }
            pGSM = (gsm_byte *) pRtp->pData + RTP_HDR_SIZE;
            DataLen = pRtp->Len - RTP_HDR_SIZE;
            DecompressGSM(pGSM,DataLen,gsmh);
            break;

         case RTP_PT_DVI4_8K:
            sp = (u_int8 *) pRtp->pData + RTP_HDR_SIZE;
            dp = (char *) sp  + 4;
            DataLen = pRtp->Len - RTP_HDR_SIZE - 4;
            adpcm.valprev = (sp[0] << 8) | sp[1];
            adpcm.index = sp[2];
            DecompressADPCM(dp,DataLen,&adpcm);
            break;

         default:
            break;
      }

      pNext = pRtp->Link;
      free(pRtp->pData);
      free(pRtp);
      pRtp = pNext;
   }
}


void VoipNode::DecompressAudio()
{
   if(mCC->Proto == PROTO_ILINK || mCC->Proto == PROTO_RTP) {
   // Currently the only supported compression technique is GSM
      DecompressRTP();
   }
   else if(mCC->Proto == PROTO_SF) {
   // Speak Freely doesn't have sequence numbers so any received data
   // has already been decompressed.  No need to do anything here.
   }
   else {
   // This shouldn't happen !
      CORE_DUMP();
   }
}


// Called when VoIP stream stops transmitting
void VoipNode::StopTalking()
{
   RtpData *pRtp;
   u_int16 NextSeqAvailable;
   
   if(pRtpQueue != NULL) {
   // flush everything that's left to audio
      while((pRtp = pRtpQueue) != NULL) {
      // There's uncompressed RTP data available
         NextSeqAvailable = pRtp->pData->seq;
         while(NextSeq != NextSeqAvailable) {
            D2PRINTF(("m1 %d",NextSeq));
            NextSeq++;
            MissedPackets++;
         }
         D2PRINTF(("\n"));
      // Make some audio
         DecompressAudio();
      }
   }

// Back to idle state
   bRxIdle = TRUE;
   bRxBuffering = bRxActive = FALSE;
}

void VoipNode::InsertInRTPQ(RtpData *pRtpData,int MaxNetQueueLen)
{
   RtpData *pLast = pRtpQueue;
   u_int16 Seq = pRtpData->pData->seq;
   
   if(pLast == NULL) {
   // The Queue is empty
      pLast = (RtpData *) &pRtpQueue;
   }
   else {
      RtpData *p = pLast->Link;

      if(p == NULL) {
      // only one packet on the queue
         if(Seq == pLast->pData->seq) {
            DuplicatePackets++;
            free(pRtpData->pData);
            free(pRtpData);
            pLast = NULL;
         }
         else if(Between(Seq,pLast->pData->seq,pLast->pData->seq + MaxNetQueueLen)) {
         // this packet is now the first on the queue
            D2PRINTF(("\ns(%d,%d)",pLast->pData->seq,Seq));
            OutofSeqPackets++;
            pRtpData->Link = pRtpQueue;
            pLast = (RtpData *) &pRtpQueue;
         }
      }
      else  {
         while(p != NULL) {
            if(Seq == pLast->pData->seq) {
               DuplicatePackets++;
               free(pRtpData->pData);
               free(pRtpData);
               pLast = NULL;
               break;
            }
            else if(Between(pLast->pData->seq,Seq,p->pData->seq)) {
            // insert new packet here
               pRtpData->Link = p;
               OutofSeqPackets++;
               D2PRINTF(("\ns(%d,%d,%d)",pLast->pData->seq,Seq,p->pData->seq));
               break;
            }
            pLast = p;
            p = p->Link;
         }
      }
   }

   if(pLast != NULL) {
      D2PRINTF(("q"));
      pLast->Link = pRtpData;
      RtpQCount++;
   }
}

void VoipNode::ConsumeAudio(NetAudio *pAudio)
{
   if(--pAudio->Users == 0) {
      AudioQCount--;
      ALOG(("%s Freeing BufNum %d, AudioQCount: %d\n",NodeName,pAudio->BufNum,
            AudioQCount));
      D2PRINTF(("Buffer emptied\n"));
      pInHead = pAudio->Link;
      free(pAudio);
   }
}

int VoipNode::DeleteTx(LinkedNode *pDeleteLink)
{
   int Ret = Node::DeleteTx(pDeleteLink);

// Decrement the usage count for all buffers this link hasn't used yet
   NetAudio *pAudio = pDeleteLink->pInHead;
   NetAudio *pNextAudio;
   while(pAudio != NULL) {
      pNextAudio = pAudio->Link;
      ConsumeAudio(pAudio);
      pAudio = pNextAudio;
   }
   return Ret;
}

void VoipNode::AddTx(LinkedNode *pLink)
{
   Node::AddTx(pLink);

// Increment the usage count for all queued buffers
   NetAudio *pAudio = pLink->pInHead;
   while(pAudio != NULL) {
      pAudio->Users++;
      pAudio = pAudio->Link;
   }
}

// Advance the CtoneTimer by the amount of time it will take to consume
// the specified number of bytes
void VoipNode::Add2CtoneTimer(int AudioBytes)
{
   int Duration = (AudioBytes * 500) / 8000; // in milliseconds

   int microseconds = CtoneTimer.tv_usec + ((Duration % 1000) * 1000);
   CtoneTimer.tv_usec = microseconds % 1000000;
   CtoneTimer.tv_sec  = CtoneTimer.tv_sec + (Duration / 1000) + 
                        microseconds / 1000000;
}

// Returns number of samples available at *pAudio
int VoipNode::GetRxAudio(LinkedNode *pLink,int16 **AudioOut)
{
   int CopyCount;
   RtpData *pRtp;
   u_int16 NextSeqAvailable;
   NetAudio *pAudio;

   AudioAvailable = 0;
   *AudioOut = AudioBuf;
   while(AudioAvailable < sizeof(AudioBuf)) {
   // need more audio in the output buffer
      if(pLink->pInHead == NULL && pInHead != NULL && pLink->BufNum < BufNum) {
         pAudio = pInHead;
         while(pAudio->BufNum <= pLink->BufNum) {
         // We've already consumed this buffer, try the next one
            ALOG(("%s Skipping BufNum %d already consumed by %s\n",
                  NodeName,pAudio->BufNum,pLink->pTx->NodeName));
            if((pAudio = pAudio->Link) == NULL) {
               break;
            }
         }
         if(pAudio != NULL) {
            pLink->pInHead = pAudio;
            pLink->Used = 0;
            pLink->BufNum = pAudio->BufNum;
         }
      }
      pAudio = pLink->pInHead;

      if(pAudio != NULL) {
         CopyCount = sizeof(AudioBuf) - AudioAvailable;
         CopyCount = min(CopyCount,pAudio->Available - pLink->Used);
         D2PRINTF(("Copying %d bytes to AudioBuf[%d]\n",CopyCount,AudioAvailable));
         ALOG(("%s copying %d bytes from BufNum %d for %s\n",NodeName,CopyCount,
               pAudio->BufNum,pLink->pTx->NodeName));

         memcpy(&AudioBuf[AudioAvailable/2],&pAudio->Buf[pLink->Used],CopyCount);
         AudioAvailable += CopyCount;
         pLink->Used += CopyCount;
         if(pLink->Used == pAudio->Available) {
         // Finished with this buffer
            pLink->pInHead = pAudio->Link;
            pLink->BufNum = pAudio->BufNum;
            pLink->Used = 0;
            ALOG(("%s %s consumed all of BufNum %d\n",NodeName,
                  pLink->pTx->NodeName,pAudio->BufNum));
            ConsumeAudio(pAudio);
         }
      }
      else if(bRxActive && (pRtp = pRtpQueue) != NULL) {
      // We're not buffering and there's uncompressed RTP data available
         NextSeqAvailable = pRtp->pData->seq;
         while(NextSeq != NextSeqAvailable) {
            D2PRINTF(("m2 %d ",NextSeq));
            NextSeq++;
            MissedPackets++;
         }
         D2PRINTF(("\n"));
      // Try to decompress some audio
         ALOG(("%s Missed packet\n",NodeName));
         DecompressAudio();
      }
      else if(bRxActive) {
      // Nothing more we can do, we're out out data.  
      // Go back to buffering state and get some more data
         bRxActive = FALSE;
         bRxBuffering = TRUE;
         ALOG(("%s buffer ran dry, start buffering\n",NodeName));
         D2PRINTF(("O"));
         break;
      }
      else {
         break;
      }
   }

   if(AudioAvailable > 0 && Ctone != 0 && CtoneTimer.tv_sec == 0) {
   // start the CtoneTimer to go off after the audio we have available
   // is consumed
      CtoneTimer = TimeNow;
      Add2CtoneTimer(AudioAvailable);
      if((pAudio = pLink->pInHead) != NULL) {
         Add2CtoneTimer(pAudio->Available - pLink->Used);
         pAudio = pAudio->Link;
         while(pAudio != NULL) {
            Add2CtoneTimer(pAudio->Available);
            pAudio = pAudio->Link;
         }
      }
   }

   if(AudioAvailable == 0 && CtoneTimer.tv_sec != 0 && Ctone != 0) {
   // No audio was available, check if it's time to generate a CTone
      if(TimeLapse(&CtoneTimer) >= CtoneDelay) {
         CtoneTimer.tv_sec = 0;
         SendCtone(TRUE);
      }
   }

// return number of samples available
   return AudioAvailable/2;
}

//
// RxState  
void VoipNode::QueueRTPData(RtpData *pRtp)
{
   u_int16 Seq = pRtp->pData->seq;
   int NetQueueLen = 0;
   int MaxNetQueueLen;

   switch(pRtp->pData->pt) {
      case RTP_PT_GSM:
         NetQueueLen = VoipDejitterDelay / 80;
         break;

      case RTP_PT_DVI4_8K:
         NetQueueLen = VoipDejitterDelay / 122;
         break;

      default:
         break;
   }
   MaxNetQueueLen = NetQueueLen * 2;

   if(bRxIdle) {
   // First packet received, assume this is the first sequence #
      D2PRINTF(("t%d",Seq));
      NextSeq = Seq;
      bRxIdle = FALSE;
      bRxBuffering = TRUE;
   }

   if(bRxBuffering) {
      if(Between(Seq,NextSeq,Seq + MaxNetQueueLen)) {
      // This packet is before the current NextSeq, backup NextSeq
         NextSeq = Seq;
      }
   }

   if(Between(NextSeq,Seq,NextSeq + MaxNetQueueLen)) {
      InsertInRTPQ(pRtp,MaxNetQueueLen);
      if(bRxBuffering) {
         if(RtpQCount >= NetQueueLen) {
         // We're buffering and it's time to rock and roll
            D2PRINTF(("S"));
            bRxBuffering = FALSE;
            bRxActive = TRUE;
            DecompressAudio();
         }
      }
      else if(bRxActive) {
      // try to make some more audio
         DecompressAudio();
      }
      else {
      // this shouldn't happen
         CORE_DUMP();
      }
   }
   else {
   // This packet is out of range, must have been delayed too long,
   // toss it
      D2PRINTF(("\nl(%d,%d,%d)",NextSeq,Seq,NextSeq+MaxNetQueueLen));
      LatePackets++;
      free(pRtp->pData);
      free(pRtp);
   }
}

void VoipNode::QueueSFData(soundbuf *p)
{
   int compression = (int) htonl(p->compression);
   char *pData;
   int Len = htonl(p->buffer.buffer_len);
   struct adpcm_state adpcm;      /* ADPCM compression state */
   u_int8 *sp;
   int NetQueueLen = 0;

   if(bRxIdle) {
   // First packet received, assume this is the first sequence #
      D2PRINTF(("t"));
      bRxIdle = FALSE;
      bRxBuffering = TRUE;
   }

// Speak Freely doesn't have sequence numbers so decompress data now
   switch(compression) {
      case fCompGSM | fProtocol:
         if(gsmh == NULL) {
            gsmh = gsm_create();
         }
         Len -= 2;
         pData = &p->buffer.buffer_val[2];
         NetQueueLen = VoipDejitterDelay / 200;
         DecompressGSM((gsm_byte *) pData,Len,gsmh);
         break;

      case fCompADPCM | fProtocol:
       // Restore the decoder state from the copy saved in the 
       // packet, independent of the byte order of the machine 
       // we're running on. 
         pData = p->buffer.buffer_val;
         Len -= 3;
         sp = (u_int8 *) &pData[Len];
         adpcm.valprev = (sp[0] << 8) | sp[1];
         adpcm.index = sp[2];
         NetQueueLen = VoipDejitterDelay / 122;
         DecompressADPCM(pData,Len,&adpcm);
         break;

      default:
         CORE_DUMP();
         break;
   }

   if(bRxBuffering) {
      if(AudioQCount >= NetQueueLen) {
      // We're buffering and it's time to rock and roll
         D2PRINTF(("S"));
         bRxBuffering = FALSE;
         bRxActive = TRUE;
      }
      else {
         D2PRINTF(("q"));
      }
   }
   else if(!bRxActive) {
   // this shouldn't happen
      CORE_DUMP();
   }
}


void VoipNode::EndPointCleanup()
{
   NetAudio *pAudio;
   RtpData *pRtp;

   if(gsmh != NULL) {
      gsm_destroy(gsmh);
   }

   while((pAudio = pInHead) != NULL) {
      pInHead = pAudio->Link;
      AudioQCount--;
      free(pAudio);
   }

   while((pRtp = pRtpQueue) != NULL) {
      pRtpQueue = pRtp->Link;
      free(pRtp->pData);
      free(pRtp);
   }
}

void EndPointCleanup(ConfClient *pCC)
{
   VoipNode *pVoipNode = (VoipNode *) pCC->pVoipNode;
   if(pVoipNode != NULL) {
      pVoipNode->EndPointCleanup();
      delete pVoipNode;
   }
}

VoipNode::~VoipNode()
{
   if(CurrentNode == this) {
   // Set the CurrentNode to *Something*
      struct avl_traverser avl_trans;
      CurrentNode = (Node *) avl_t_first(&avl_trans,NodeTree);
   }

   if(avl_delete(NodeTree,this) == NULL) {
      LOG_ERROR(("%s: avl_delete() failed to find me.\n",__FUNCTION__));
   }
}

int VoipNode::WriteAudioOut(ClientInfo *p)
{
   LOG_ERROR(("%s: I should never be called!.\n",__FUNCTION__));
   CORE_DUMP();
   return 0;
}

void VoipNode::GsmTx()
{
   gsm_byte *pGSM;
   int i;
   rtp_hdr_t *pRtpHdr;

   pGSM = (gsm_byte *) &GSM_Packet[SIZEOF_SOUNDBUF+2];
   pGSM += GSM_FRAME_LEN * GsmFramesAvail;

// Compress the audio

   for(i = 0; i < AUDIO_BUF_SIZE; i += AUDIO_SAMPLES_PER_GSM_FRAME) {
      gsm_encode(hGSM_Encode,&AudioBuf[i],pGSM);
      pGSM += sizeof(gsm_frame);
   }
   pGSM = (gsm_byte *) &GSM_Packet[SIZEOF_SOUNDBUF+2];
   pGSM += GSM_FRAME_LEN * GsmFramesAvail;

   if(mCC->Proto == PROTO_RTP || mCC->Proto == PROTO_ILINK) {
   // Send Echolink or RTP packets

      pRtpHdr = (rtp_hdr_t *) (((char *) pGSM) - RTP_HDR_SIZE);
      memset(pRtpHdr,0,RTP_HDR_SIZE);
      pRtpHdr->pt = RTP_PT_GSM;
      pRtpHdr->seq = htons(TxSequence++);
      pRtpHdr->ssrc = OurNodeID;

      pRtpHdr->version = mCC->Proto == PROTO_ILINK ? ILINK_RTP_VERSION : 2;
      SendAudioPacket(pRtpHdr,ILINK_RTP_FRAME_LEN);
   }
   else if(mCC->Proto == PROTO_SF) {
   // Send SF frames
      do {
      // Fill in the Send Speak Freely header
         soundbuf *pSFHdr = (soundbuf *) GSM_Packet;
         int TxLen;
   
         GsmFramesAvail += AUDIO_BUF_SIZE / AUDIO_SAMPLES_PER_GSM_FRAME;
         if(GsmFramesAvail < 10) {
         // Not enough data yet
            break;
         }
         GsmFramesAvail -= 10;
   
      // Hide the original number of samples in first 2 bytes of buffer.
         *((short *) pSFHdr->buffer.buffer_val) = 
                                    htons(AUDIO_SAMPLES_PER_GSM_FRAME*10);
   
         pSFHdr->compression = (int) htonl(fCompGSM | fProtocol);
         pSFHdr->buffer.buffer_len = (int) htonl(2 + (10*GSM_FRAME_LEN));
         TxLen = SIZEOF_SOUNDBUF + 2 + (10*GSM_FRAME_LEN);
   
         memset(pSFHdr->sendinghost,0,sizeof(pSFHdr->sendinghost));
         strncpy(pSFHdr->sendinghost,ConferenceCall,sizeof(pSFHdr->sendinghost));
   
      // Send Speak Freely ADPCM packets
   
         SendAudioPacket(pSFHdr,TxLen);
   
         if(GsmFramesAvail > 0) {
         // Copy GSM data we didn't use to start of buffer
            char *pTo = &GSM_Packet[SIZEOF_SOUNDBUF+2];
            char *pFrom = pTo + (GSM_FRAME_LEN * 10);
            memcpy(pTo,pFrom,GsmFramesAvail * GSM_FRAME_LEN);
         }
      } while(FALSE);
   }
   else {
      LOG_ERROR(("%s: Error protocol = %d.\n",__FUNCTION__,mCC->Proto));
   }
}

void VoipNode::SendAudioPacket(void *Buf,int TxLen)
{
#if 0
   if(ConfEnable && mCC->bInConf) {
   // Conferencing is enabled and this client is in the conference,
   // let the conference code send the data
   }
   else {
#endif
   // Just send it to him directly
      Send2(mCC,Buf,TxLen,mCC->pCS->AudioPort);
// }
}

void VoipNode::AdpcmTx()
{
   int Samples2Encode;
   int SamplesUsed = 0;

   do {
      Samples2Encode = (ADPCM_PACKET_DATA_BYTES - ADPCM_Bytes) * 2;
      if(Samples2Encode > (AUDIO_BUF_SIZE - SamplesUsed)) {
         Samples2Encode = AUDIO_BUF_SIZE - SamplesUsed;
      }
#if 0
      printf("Encode %d samples from %d to %d\n",
             Samples2Encode,SamplesUsed,ADPCM_Bytes);
#endif
      adpcm_coder(&AudioBuf[SamplesUsed],&Adpcm.Buf[ADPCM_Bytes],
                  Samples2Encode,&TxADPCM);
      ADPCM_Bytes += Samples2Encode/2;
      SamplesUsed += Samples2Encode;

      if(ADPCM_Bytes == ADPCM_PACKET_DATA_BYTES) {
      // The ADPCM packet is full, sendit
         ADPCM_Bytes = 0;

//       printf("packet full, sendit\n");

         if(mCC->Proto == PROTO_SF) {
         // Fill in the Send Speak Freely header
            soundbuf *pSFHdr = (soundbuf *) &Adpcm.Buf[-SIZEOF_SOUNDBUF];
            u_int8 *pData = (u_int8 *) Adpcm.Buf;
            u_int8 *sp = (u_int8 *) &pData[ADPCM_PACKET_DATA_BYTES];
            int TxLen = SIZEOF_SOUNDBUF + ADPCM_PACKET_DATA_BYTES + 3;

            pSFHdr->compression = (int) htonl(fCompADPCM| fProtocol);
            pSFHdr->buffer.buffer_len = (int) htonl(ADPCM_PACKET_DATA_BYTES+3);
            memset(pSFHdr->sendinghost,0,sizeof(pSFHdr->sendinghost));
            strncpy(pSFHdr->sendinghost,ConferenceCall,
                    sizeof(pSFHdr->sendinghost));
            sp[0] = (u_int8) (((u_int16)TxADPCMStart.valprev) >> 8);
            sp[1] = (TxADPCMStart.valprev & 0xff);
            sp[2] = TxADPCMStart.index;

         // Send Speak Freely ADPCM packets
            SendAudioPacket(pSFHdr,TxLen);
         }
         else if(mCC->Proto == PROTO_RTP) {
         // Fill in the RTP header
            rtp_hdr_t *pRtpHdr = (rtp_hdr_t *) &Adpcm.Buf[-(RTP_HDR_SIZE+4)];
            u_int8 *sp = (u_int8 *) &Adpcm.Buf[-4];
            int TxLen = RTP_HDR_SIZE + ADPCM_PACKET_DATA_BYTES + 4;

            memset(pRtpHdr,0,sizeof(rtp_hdr_t));
            pRtpHdr->pt = RTP_PT_DVI4_8K;
            pRtpHdr->seq = htons(RtpTxSequence++);
            pRtpHdr->ssrc = OurNodeID;
            pRtpHdr->version = 2;
            sp[0] = (u_int8) (((u_int16)TxADPCMStart.valprev) >> 8);
            sp[1] = (TxADPCMStart.valprev & 0xff);
            sp[2] = TxADPCMStart.index;

         // Send RTP ADPCM packets
            SendAudioPacket(pRtpHdr,TxLen);
         }
         else {
            LOG_ERROR(("%s: Error protocol = %d.\n",__FUNCTION__,mCC->Proto));
         }
         TxADPCMStart = TxADPCM;
      }
   } while(SamplesUsed < AUDIO_BUF_SIZE);
}

void VoipNode::PollTx()
{
   GetTxAudio();
   if(AudioAvailable > 0) {
#if 0
      if(fp != NULL) {
         fwrite(AudioBuf,AudioAvailable,1,fp);
      }
#endif
      if(mCC->CompressionType == RTP_PT_DVI4_8K) { 
         AdpcmTx();
      }
      else if(mCC->CompressionType == RTP_PT_GSM) {
         if(hGSM_Encode == NULL) {
            if((hGSM_Encode = gsm_create()) == NULL) {
               LOG_ERROR(("%s: gsm_create failed.\n",__FUNCTION__));
            }
         }
         
         if(hGSM_Encode != NULL) {
            GsmTx();
         }
      }
      else {
         LOG_ERROR(("%s: Error unknown CompressionType %d.\n",__FUNCTION__,
                    mCC->CompressionType));
      }
   }

   CheckDtmfTimeouts();
}

extern struct config_entry VoipConfigVars[];

void VoipNode::VoipInit(ConfClient *pCC)
{
   Node *CurrentSave = CurrentNode; // save the current node pointer
   Node *pNode = NULL;

   char *ConfigFile[2] = {
      NULL,
      "voip.conf"
   };

   do {
      pNode = FindNode(pCC->Callsign);
      if(pNode != NULL) {
      // Hmmm... the node already exists !
         LOG_ERROR(("%s: node %s already exists!\n",__FUNCTION__,pCC->Callsign));
         break;
      }

      if(pCC->pVoipNode != NULL) {
         LOG_ERROR(("%s: pVoipNode != NULL!\n",__FUNCTION__));
         break;
      }

      if((pNode = new VoipNode(pCC)) == NULL) {
         LOG_ERROR(("%s: new failed\n",__FUNCTION__));
         break;
      }
      pCC->pVoipNode = CurrentNode = pNode;

   // If there is a configuration file <callsign>.conf or
   // voip.conf then use it to configure the new voip node

      if((ConfigFile[0] = (char *) malloc(strlen(pCC->Callsign)+6)) == NULL) {
         LOG_ERROR(("%s: malloc failed\n",__FUNCTION__));
         break;
      }
      char *cp = ConfigFile[0];
      strcpy(cp,pCC->Callsign);

      while(*cp) {
         *cp = tolower(*cp);
         cp++;
      }
      strcpy(cp,".conf");

      for(int i = 0; i < 2; i++) {
         FILE *fp;
         if((fp = fopen(ConfigFile[i],"r")) != NULL) {
            LOG_NORM(("%s: Configuring new VoIP client from %s\n",
                      __FUNCTION__,ConfigFile[i]));

            for(int Pass = 1; Pass < 3; Pass++) {
               int Err;
               if((Err = read_config(fp,VoipConfigVars,Pass)) != 0) {
                  LOG_ERROR(("%s: read_config failed (%d)\n",__FUNCTION__,Err));
               }
               fseek(fp,0,SEEK_SET);
            }
            fclose(fp);
            break;
         }
      }

      if(CurrentNode->DtmfMethod == 1) {
      // Software DTMF decoder enabled, initialize it
         dtmf_rx_init(&pNode->DtmfRxState,NULL,pNode);
         dtmf_rx_set_realtime_callback(&pNode->DtmfRxState,DtmfCallback,pNode);
      }
   } while(FALSE);

   if(pNode != NULL) do {
      char *ConnectMap = NULL;

      if(!pCC->bPermanent && VoipInMap != NULL) {
      // inbound connection, connect this client to the specified ports
         ConnectMap = VoipInMap;
      }

      if(pCC->bPermanent && VoipOutMap != NULL) {
      // outbound connection, connect this client to the specified ports
         ConnectMap = VoipOutMap;
      }

      if(ConnectMap != NULL) {
      // Initial connection map found, setup the links
         char *cp = ConnectMap;
         char *cp1;
         Node *pNode2Link;
         char CharSave = 0;

         while(*cp) {
            if((cp1 = strchr(cp,' ')) != NULL ||
               (cp1 = strchr(cp,' ')) != NULL) 
            {
               CharSave = *cp1;
               *cp1++ = 0;
            }
            if((pNode2Link = FindNode(cp)) != NULL) {
            // Form link
               pNode2Link->AddRx(pNode);
               pNode->AddRx(pNode2Link);
            }
            else {
               LOG_ERROR(("%s: %s node %s not found.\n",__FUNCTION__,
                          pCC->bPermanent ? "VoipOutMap" : "VoipInMap",cp));
            }
            if(cp1 != NULL) {
               cp1[-1] = CharSave;
               cp = cp1;
               while(*cp == ' ') {
                  cp++;
               }
            }
            else {
               break;
            }
         }
      }
   } while(FALSE);

   if(ConfigFile[0] != NULL) {
      free(ConfigFile[0]);
   }

   CurrentNode = CurrentSave;
}

int VoipNode::Enabled()
{
   int Ret = TRUE;

   if(!mCC->bInConf && mCC->bSWL && mCC->bKicked) {
   // In the disconnecting state
      Ret = FALSE;
   }
   return Ret;
}


